/*
   This file is part of msplot, a visualisation tool for Massif output

   Copyright (C) 2007-2008 Martijn Versteegh


   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.

   The GNU General Public License is contained in the file COPYING.
*/

#ifndef __PARSE_H__
#define __PARSE_H__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// parse.cpp and parse.h implement the datastructures to hold a massif session, and the routines to parse it from a massif output file



// TreeNode trees are used to store massif's callback trees
class TreeNode
{
	public:
		TreeNode(int _size, char const *_name, bool insig)
		{
			parent = NULL;
			size = _size;
			name = strdup(_name);
			numChildren = maxNumChildren = 0;
			children = NULL;
			id = 0;
			isInsignificant = insig;
		}
		~TreeNode()
		{
			delete [] children;
			free(name);
		}
		
		void AddChild(TreeNode *c)
		{
			if (numChildren >=  maxNumChildren)
				Grow();

			children[numChildren++] = c;
			c->parent = this;
		}

		//GetId() returns a unique id for each treenode, which is gueranteed to be the same for
		//different snapshots. It returns 0 before Enumerate is called.
		unsigned int GetId() const { return id; }

		TreeNode const *GetParent() const { return parent; }

		TreeNode const *FindNode(unsigned int idToFind) const
		{
			if (id == idToFind)
			{
				return this;
			}

			TreeNode const *ret;
			for (int i = 0; i < numChildren; i++)
			{
				ret = children[i]->FindNode(idToFind);

				if (ret)
					return ret;
			}

			return NULL;
		}

		// clone this node and it's children
		TreeNode *Clone() const
		{
			TreeNode *ret = new TreeNode(size, name, isInsignificant);
			for (int i = 0; i < numChildren; i++)
			{
				ret->AddChild(children[i]->Clone());
			}

			return ret;
		}

		// enumerate the tree, this generates id's for all the children
		void Enumerate(unsigned int *seed)
		{
			id = (*seed)++;

			for (int i = 0; i < numChildren ; i++)
			{
				children[i]->Enumerate(seed);
			}
		}

		// look in t's children for a node with the same name, take it's id, and enumerate our children from it
		//  each session stores a global tree, which contains all nodes existing in all snapshots
		//  by enumreating each snapshot from this gloabl tree we make sure the id's stay the same for each snapshot
		void EnumerateFrom(TreeNode const *t)
		{
			for (int i = 0 ; i < t->numChildren ; i++)
			{
				if (!strcmp(name, t->children[i]->name))
				{
					id = t->children[i]->id;

					for (int j = 0; j < numChildren; j++)
					{
						children[j]->EnumerateFrom(t->children[i]);
					}
				}
			}
		}

		// this is used to build up the global tree
		//  it checks if this node already contains a child of the same name, and if so recursively merges the children
		// or if it doen's yet exist it clones them
		void MergeChild(TreeNode const *t);

		unsigned int GetNumChildren() const { return numChildren; }
		TreeNode *const GetChild(int i) const { return children[i]; }
		char const *GetName() const { return name; }

		unsigned int GetSize() const { return size; }

		void DumpXML(int leadingSpaces)
		{
			for (int i = 0; i < leadingSpaces; i++)
				printf(" ");

			printf("<treenode memsize=\"%d\" name=\"%s\">\n", size, name);
			for (int i = 0; i < numChildren; i++)
			{
				children[i]->DumpXML(leadingSpaces+1);
			}
			
			for (int i = 0; i < leadingSpaces; i++)
				printf(" ");
			printf("</treenode>\n");
			
		}


		void FindMaxDepth(bool skipSingle, unsigned int level, unsigned int *depth) const
		{
			if (level > *depth)
				*depth = level;

			if (skipSingle && numChildren < 2)
				return;
				
			for (int i = 0; i < numChildren; i++)
			{
				children[i]->FindMaxDepth(skipSingle, level +1, depth);
			}

		}
		
	private:
		void Grow()
		{
			if (children)
			{
				maxNumChildren *=2;
				TreeNode **nw = new TreeNode *[maxNumChildren];
				for (int i = 0; i< numChildren ;i++)
				{
					nw[i] = children[i];
				}
				delete [] children;
				children =nw;
			}
			else
			{
				maxNumChildren = 1;
				children = new TreeNode *[maxNumChildren];
			}
		}
		int numChildren;
		int maxNumChildren;
		TreeNode **children;

		unsigned int id;

		int size;
		char *name;
		TreeNode *parent;
		bool isInsignificant;
};


// a Snapshot holds all information of a single massif snapshot
class SnapShot
{
	public:
		SnapShot(unsigned int _nr)
		{
			nr = _nr;
			time = 0;
			heap = 0;
			heap_extra = 0;
			stack = 0;
			allocTree = NULL;
			dummy = new TreeNode(0, "no tree available", true);
		}

		void SetTime(unsigned int t)
		{
			time = t;
		}

		void SetHeap(unsigned int h)
		{
			heap = h;
			if (dummy)
				delete dummy;
			dummy = new TreeNode(heap, "no tree available", true);
		}

		void SetHeapExtra(unsigned int h)
		{
			heap_extra = h;
		}

		void SetStack(unsigned int s)
		{
			stack = s;
		}

		void AddAllocTree(TreeNode *t)
		{
			allocTree = t;
		}

		
		~SnapShot()
		{
			if (allocTree)
				delete allocTree;

			delete dummy;
		}

		TreeNode const *GetAllocTree() const
		{
			if (allocTree)
				return allocTree;


			return dummy;
		}

		void DumpXML()
		{
			printf(" <snapshot nr=\"%d\" time=\"%d\" heap=\"%d\" heap_extra=\"%d\" stack=\"%d\" alloctree=\"%s\">\n", nr, time, heap, heap_extra, stack, allocTree ? "detailed" : "empty" );

			if (allocTree)
				allocTree->DumpXML(2);
			printf(" </snapshot>\n");
		}

		unsigned int GetMaxDepth(bool ignoreSingleChildren)
		{
			if (!allocTree)
				return 0;

			unsigned int depth = 0;

			allocTree->FindMaxDepth(ignoreSingleChildren, 1, &depth);

			return depth;
		}


		void EnumerateFrom(TreeNode const *from)
		{
			if (allocTree)
			{
				allocTree->EnumerateFrom(from);
			}
		}


		TreeNode const *FindNode(unsigned int id)
		{
			TreeNode const *ret = NULL;
			if (allocTree)
			{
				ret = allocTree->FindNode(id);
			}

			return ret ? ret : dummy;
		}

	private:

		unsigned int nr, time, heap, heap_extra, stack;
		TreeNode *allocTree;
		TreeNode *dummy;

};

// this holds all information of a massif outputfile
class Session
{
	public:
	Session()
		: collectedTree(0,"top", true)
	{
		desc = NULL;
		cmd = NULL;
		time_unit = NULL;

		snapShots = NULL;
		numSnapShots = maxNumSnapShots = 0;
	}

	void SetDesc(char const *s)
	{
		desc = strdup(s);
	}

	void SetCmd(char const *s)
	{
		cmd = strdup(s);
	}

	void SetTimeUnit(char const *s)
	{
		time_unit = strdup(s);
	}

	~Session()
	{
		free(desc);
		free(time_unit);
		free(cmd);

		if (snapShots)
		{
			for (int  i =0; i < numSnapShots; i++)
			{
				delete snapShots[i];
			}
			
			delete [] snapShots;
		}
	}

	void AddSnapShot(SnapShot *s)
	{
		if (numSnapShots >= maxNumSnapShots)
			Grow();

		snapShots[numSnapShots++] = s;

		collectedTree.MergeChild(s->GetAllocTree());
	}


	void DumpXML()
	{
		printf("<session cmd=\"%s\" desc=\"%s\" time_unit=\"%s\">\n", cmd, desc, time_unit);
		for (int i = 0; i < numSnapShots; i++)
		{
			snapShots[i]->DumpXML();
		}
		printf("</session>");
	}


	SnapShot const * GetSnapShot(unsigned int s)
	{
		return snapShots[s];
	}

	unsigned GetNumSnapShots() { return numSnapShots; }


	unsigned int GetMaxSize()
	{
		unsigned int ret = 0, s = 0;
		for (int i = 0; i < numSnapShots; i++)
		{
			s = snapShots[i]->GetAllocTree()->GetSize();

			if (s > ret)
				ret = s;
		}

		return ret;
	}
	

	unsigned int GetMaxDepth(bool ignoreSingleChildren = true)
	{
		unsigned int ret = 0, l;
		for (int i = 0; i < numSnapShots; i++)
		{
			l = snapShots[i]->GetMaxDepth(ignoreSingleChildren);

			if (l > ret)
				ret = l;
		}

		return ret;
	}


	void DoFinalBookKeeping() // called after the file has been parsed to scan the tree
	{
		// calc maxmem and maxlevel as well here?
		unsigned int enumerator = 1;
		collectedTree.Enumerate(&enumerator);

		for (int i = 0; i < numSnapShots ; i++)
		{
			snapShots[i]->EnumerateFrom(&collectedTree);
		}

//		collectedTree.DumpXML(0);
	}

	TreeNode const *FindNode(unsigned int snapShot, unsigned int id)
	{
		return snapShots[snapShot]->FindNode(id);
	}
	
	private:

	char *desc, *cmd, *time_unit;
	int numSnapShots, maxNumSnapShots;
	void Grow()
	{
		if (!maxNumSnapShots)
		{
			maxNumSnapShots = 64;
		}
		else
		{
			maxNumSnapShots *= 2;
		}

		SnapShot **tmp = new SnapShot *[maxNumSnapShots];

		for (int i = 0; i < numSnapShots; i++)
			tmp[i] = snapShots[i];

		delete [] snapShots;

		snapShots = tmp;
	}

	SnapShot ** snapShots;

	TreeNode collectedTree; // this tree is build up by merging all snapshot trees, and thus contains all nodes that exist in the log, it is used for enumeration of the nodes



	
};




extern Session *ReadMassifFile(FILE *in);

#endif
